function out = Channels2NCC (channels, radii, nangles, beta)
% Channels2NCC : This function takes each channel of the image and
% considers various scales and angles. The final output is a cell array
% with nangles entries each entry is an image with nchannels*nscales
% entries.

channels_2 = channels.^2;

nrows     = size(channels,1);
ncols     = size(channels,2);
nchannels = size(channels,3);


% Preallocate output
out = zeros(nangles, nrows, ncols, nchannels*length(radii), 'single');

idx = 0;

% For each scale
for radius = radii
    
    [x, y] = meshgrid(-radius:radius, -radius:radius);
    
    % Figure out which pixels are inside the circle
    circle_mask = double ( (x.^2 + y.^2) <= (radius*radius) );
    
    n = nnz (circle_mask);
    
    % Compute the variances in each circular window for each channel
    scale = zeros(nrows, ncols, nchannels);
    
    for i = 1:nchannels
        SumX  = imfilter (channels(:,:,i),  circle_mask, 'replicate');
        SumX2 = imfilter (channels_2(:,:,i), circle_mask, 'replicate');
        
        % Computing the window_norm
        window_norm = sqrt (max(0, SumX2 - (SumX.^2/n)));
        
        scale(:,:,i) = 1 ./ ( beta*mean(window_norm(:)) + window_norm );
    end
    
    % For each orientation
    for i = 1:nangles
        
        angle = deg2rad( (i-1)*(180.0/nangles));
        
        c = cos(angle);
        s = sin(angle);
        
        %% Produce oriented edge template
        x2 = c*x - s*y;
        
        % Here we just use a sigmoid along the x direction
        template = ( x2 ./ sqrt(1 + x2.^2) ) .* circle_mask;
        
        % Normalize the sum to zero mean
        template = template - mean(template(:));
        
        % Normalize to unit norm
        template = template / norm(template(:));
        
        %% Compute gradients on all channels by convolving with the template
        grad = abs ( imfilter (channels, template, 'replicate') );
        
        %% Normalize the edge score result by multiplying by the scale - result should be between 0 and 1        
        out(i,:,:,idx + (1:nchannels)) =  grad .* scale;
        
    end
    
    idx = idx + nchannels;
    
end