%
% SegmentationScript (V17)
%

%% Read image

% Before calling this script the user should set the variable fname to a
% string corresponding to the name of the file she wants to analyze. The
% script takes care of all of the required operations and produces a number
% of results including ucm2 - an ultrametric contour map corresponding to
% the final segmentation.

% DEBUG
% fname = '../BSR/BSDS500/data/images/train/24004.jpg';

I = imread (fname);

nrows = size(I,1);
ncols = size(I,2);

%% Set up parameters

load EdgeParamsV6;

rho         = 25;
neigs       = 30;
sigma_l     = 400;
alpha2      = 100;

%% Get edges

tic;

% Compute channels
channels = RGB2Channels(I, kernels);

% Compute NCC outputs
NCC = Channels2NCC (channels, radii, nangles, beta);

% Get edges - highThresh = lowThresh = 0.0
[e, max_strength, max_angle, nccOE] = edgeNCCV3 (NCC, weights, [0 0]);

nccTime = toc;

%% Compute the watershed transform

% watershed transform to divide the image into superpixels
labels = watershed (max_strength);

% Creating the label matrix
L = makeL(labels, 3, 3);

%% Intervening contour - sample max_strength

[val, rowIdx, colIdx] = mxBuildWv4(double(max_strength), 5);

% Version that uses non-maximal suppression
% [val, rowIdx, colIdx] = mxBuildWv2(double(e .* max_strength), 5);

%% Compute edge weights

edge_weight = exp(-rho*val);

%% Construct the requisite matrices

tic;

W = sparse(rowIdx, colIdx, edge_weight, nrows*ncols, nrows*ncols);

% Symmetrize
W = 0.5*(W + W');

% Add diagonal components - each pixels affinity for itself
W = W + speye(nrows*ncols);

D = spdiags(full(sum(W,2)), 0, nrows*ncols, nrows*ncols);

% Project onto subspace
D2 = L'*D*L;
W2 = L'*W*L;

buildW2Time = toc;

%% Spectral analysis
opts.issym=1;
opts.isreal = 1;

tic;

[EigVect, EVal] = eigs(D2 - W2, D2, neigs+1, 'sm', opts);

eigsTime = toc;

% Sort the Eigenvalues and eigenvectors in ascending order
[evs, sort_order] = sort(full(diag(EVal)));
EigVect = EigVect(:, sort_order);

%% Construct full images

temp = L * EigVect(:,2:end);

% Compute scale factors associated with each eigenvector
scale = (max(temp) - min(temp)) .* sqrt(evs(2:end))';

temp = bsxfun (@times, temp, 1./scale);

eigChannels = reshape(temp, nrows, ncols, neigs);

%% Compute gradients
spectralOE = orientedGradients (eigChannels, nangles, 3);

%% Combine spectral and local components
totalOE = (sigma_l * nccOE) + spectralOE;

%% Rescale totalOE linearly onto the range 0-1, then permute to the order contours2ucm expects 
maxOE = max(totalOE(:));

tic;

ucm2 = contours2ucm (permute(totalOE, [2 3 1]) ./ maxOE, 'doubleSize');

contours2ucmTime = toc;

%% Rescale ucm2 using maxOE and alpha2

ucm2 = maxOE*ucm2;

ucm2 = ucm2 ./ (alpha2 + ucm2);

%% Display results

% Rasterize results
w_ncc = e .* max_strength;
w_ncc   = uint8(255*(1-w_ncc));
imtool (w_ncc);

w_final   = uint8(255*(1-ucm2));
imtool (w_final);

%% Segment ODS

% This ODS threshold depends upon the dataset - this value may not be
% optimal for your image you can change it to produce different
% segmentations.
ODS_threshold = 0.52;

% Translate the triangle labels into pixel labels
labels2 = bwlabel (ucm2 < ODS_threshold);
labels2 = imdilate(labels2, ones(3));
labels2 = labels2(3:2:end, 3:2:end);

ODS_result = colorImageSegments (I, labels2, [1 0 0]);

imtool (ODS_result);

%% Print out timing info

fprintf (1, 'Time taken to extract edges         = %12.2f seconds\n', nccTime);
fprintf (1, 'Time taken to build affinity matrix = %12.2f seconds\n', buildW2Time);
fprintf (1, 'Time taken to find %d eigenvectors  = %12.2f seconds\n', neigs, eigsTime);
fprintf (1, 'Time taken to run contours2ucm      = %12.2f seconds\n', contours2ucmTime);
