function [e, max_strength, max_angle, all_strength] = edgeNCCV3 (NCC, weights, thresh)
% edgeNCCV3 : This function implements a version of the idea in Meer's PAMI
% paper on edge detection which essentially advocates computing the
% Normalized Cross Correlation of each window with an oriented edge
% template. This idea amounts to convolving with the normalized edge
% templates and then dividing by the window variance.
% PAMI paper "Edge Detection with Embedded Confidence" Meer and Georgescu
% This version does this over a range of intensity and texture channels and
% then commbines the results via a weighting scheme.
% The final result is rescaled to 0-1 using the transform from odds to
% probability.

nangles     = size(NCC,1);
nrows       = size(NCC,2);
ncols       = size(NCC,3);
nchannels   = size(NCC,4);

%% Weight the channels producing an array of values
NCC2 = reshape(NCC(:), nangles*nrows*ncols, nchannels);

%% Compute log odds from affine model
% log_odds = reshape((NCC2*weights(1:end-1)) + weights(end), nangles, nrows, ncols);

%% Transform from log odds to probabilities 0-1
% all_strength = 1 ./ (1 + exp(-log_odds));

%% This hack assumes +ve weights - it should produce values between 0 and 1
w = weights(1:end-1) ./ sum(weights(1:end-1));

all_strength = reshape(NCC2 * w, nangles, nrows, ncols);

%% Initialize outputs
max_strength    = zeros(nrows, ncols);
max_angle       = zeros(nrows, ncols);
e               = false(nrows, ncols);

%% Main loop
for i = 1:nangles
    
    angle = deg2rad( (i-1)*(180.0/nangles) );
    
    c = cos(angle);
    s = sin(angle);

    % Pull out the response at this orientation
    strength = squeeze(all_strength(i,:,:));
    
    % Update max_strength and max_angle
    test1 = strength > max_strength;
    max_strength = max(strength, max_strength);
    max_angle(test1) = angle;
    
    % Test for local maxima
    
    % Note that we adopt an x-y coordinate system where the +ve x-axis
    % points to the right and the +ve y-axis points up as per usual.
    % Contrast this with the image coordinate frame which is indexed
    % relative to the upper left hand corner.
    
    % Central portion of the signal array
    center = strength(2:end-1, 2:end-1);
    
    local_maxima = (center > sample_offset(strength, c, s)) & (center > sample_offset(strength, -c, -s));
    
    % Pad the output with zeros so it is the same size as I
    local_maxima = padarray(local_maxima, [1 1]);

    e = (test1 & local_maxima) | (~test1 & e);

end

%% Apply Hysteresis thresholding

highThresh = thresh(1);
lowThresh  = thresh(2);

if (highThresh > 0)
    [rowIdx, colIdx] = find (max_strength > highThresh);
    
    BW = e & (max_strength >  lowThresh);
    
    e = bwselect (BW, colIdx, rowIdx, 8);
end

% apply thinning with bwmorph
e = bwmorph(e, 'thin', 1);  % Thin double (or triple) pixel wide contours

end


function out = sample_offset (I, dx, dy)
% Sample the intensity values in an array at a given offset - note we
% assume that dx and dx are between -1 and +1.

% Convention for naming variables Zxy

% Central pixels
z00 = I(2:end-1, 2:end-1);

if ((dx > 0) && (dy > 0))
    
    z10 = I(2:end-1, 3:end);
    z01 = I(1:end-2, 2:end-1);
    z11 = I(1:end-2, 3:end);
   
elseif ((dx > 0) && (dy <= 0))
    
    z10 = I(2:end-1, 3:end);
    z01 = I(3:end,   2:end-1);
    z11 = I(3:end,   3:end);
 
elseif ((dx <= 0) && (dy > 0))
    
    z10 = I(2:end-1, 1:end-2);
    z01 = I(1:end-2, 2:end-1);
    z11 = I(1:end-2, 1:end-2);

else % (dx <= 0) && (dy <= 0)
    
    z10 = I(2:end-1, 1:end-2);
    z01 = I(3:end,   2:end-1);
    z11 = I(3:end,   1:end-2);
    
end

% Solving for coefficient of z = ax + by + cxy + d
a = z10 - z00;
b = z01 - z00;
c = (z11 + z00) - (z01 + z10);
d = z00;

out = a*abs(dx) + b*abs(dy) + c*(abs(dx)*abs(dy)) + d;

end
