//
// mxBuildWv4
// Mex routine to replace the buildW function in the BSD
//


#include "mex.h"
#include <math.h>
#include <stdlib.h>

#define MSG_ID "mxBuildWv4:arg"

#define MIN(a, b) ( ((a) < (b)) ? (a) : (b) )
#define MAX(a, b) ( ((a) > (b)) ? (a) : (b) )

/*
 * This function tests whether the specified argument is a 2D double array or scalar. You can
 * use the two arguments, nrows and ncols, to test the size of the array or set them to 0 if you don't care.
 */

double *checkArgDoubleArray(const mxArray *prhs[], int arg_index, char *error_msg, int nrows, int ncols) {
    const mxArray *in = prhs[arg_index];
    
    if (!mxIsDouble(in) || mxIsComplex(in) || (mxGetNumberOfDimensions(in) != 2)) {
        mexErrMsgIdAndTxt(MSG_ID, error_msg);
    }
    
    if ((nrows >= 1) && (mxGetM(in) != nrows)) {
        mexErrMsgIdAndTxt(MSG_ID, error_msg);
    }
    
    if ((ncols >= 1) && (mxGetN(in) != ncols)) {
        mexErrMsgIdAndTxt(MSG_ID, error_msg);
    }
    
    return (mxGetPr(in));
}

// Function to check if integral coordinates, 0 indexed, are in bounds
int valid_pt (int x, int y, int nrows, int ncols)
{
    return ((x >= 0) && (x < ncols) && (y >= 0) && (y < nrows));
}


// This function proceeds in a similair manner to the buildW function in the BSD. Here we consider the edges between
// pixels as we traverse the brezenham line. Consider the set of pixels shown below.
//          A | B
//          -   -
//          C | D
// We associate the edge strength at pixel A with the boundaries to the right and the bottom of the pixel. When we walk along
// a line we consider all of the edges we may have to cross and accumulate the maximum edge strength.

// Find the maximal entry along the edge
double maximal_edge_entry(int x1, int y1, int x2, int y2, int nrows, int ncols, double *max_strength_ptr) {
    int dx = abs(x2 - x1);
    int dy = abs(y2 - y1);
    int length, i, x, y, oldx, oldy, idx;
    double strength, max_strength;
    
    length = (dx > dy) ? dx : dy;
    
    
    // Note that in this function (x1, y1) and (x2, y2) are zero indexed
    
    // Note that the two endpoints could be coincident in which case length would be zero and we wouldn't do any of the loop
    
    // Note that we are assuming that x1,y1 and x2,y2 are valid coordinates so the following loop will only sample valid pixels
    
    max_strength = 0.0;
    oldx = x1;
    oldy = y1;
    for (i = 1; i <= length; ++i) {
        
        // Here we are trying to compute raster coordinates using integer arithmetic in such a way that
        // we get the rounding correct: round (a/b) = floor ((a/b) + (1/2)) = floor ((2*a + b)/(2*b))
        // Note that this only works if a and b are POSITIVE
        
        // When (i==0) (x,y) == (x1,y1), when (i == length) (x,y) == (x2,y2)
        
        x = ( 2*(x2*i + x1*(length-i)) + length ) / (2*length);
        y = ( 2*(y2*i + y1*(length-i)) + length ) / (2*length);
        
        // Consider the transition between (oldx, oldy) and (x,y)
        
        if (oldx == x) {
            strength = max_strength_ptr[MIN(oldy,y) + (x*nrows)];
        } else if (oldy == y) {
            strength = max_strength_ptr[y + (MIN(oldx,x)*nrows)];
        } else {
            // upper left corner of the complex of 4
            idx = MIN(oldy,y) + (MIN(oldx,x)*nrows);
            // This should handle all 4 cases in one compact line
            strength = MAX(max_strength_ptr[idx], MAX(max_strength_ptr[idx+1], max_strength_ptr[idx+nrows]));
        }
        
        if (strength > max_strength) max_strength = strength;
        
        oldx = x;
        oldy = y;
    }
    
    return max_strength;
}

// Return column major index, x and y are zero indexed the output is indexed from 1
int sub2ind (x, y, nrows) {
    return ( (y + (x*nrows)) + 1 );
}

int InterveningContourStrength (int nrows, int ncols, double *max_strength_ptr, double radius,
                                double *val_ptr, double *rowIdx_ptr, double *colIdx_ptr)
{
    int x1, y1, x2, y2, dx, dy, idx;
    int r = (int) radius;
    double r2 = radius*radius;    
    
    idx = 0;
    
    for (x1=0; x1 < ncols; ++x1) {
        for (y1=0; y1 < nrows; ++y1) {
            for (dx = -r; dx <= r; ++dx) {
                for (dy = -r; dy <= r; ++dy) {
                    
                    // Check for the case where dx = dy = 0 - we don't add self connections
                    if ( (dx != 0) || (dy != 0) ) {
                        
                        x2 = x1 + dx;
                        y2 = y1 + dy;
                        
                        if (  valid_pt(x2, y2, nrows, ncols) && ((dx*dx + dy*dy) <= r2)  ) {
                            
                            // If val_ptr is NULL this is just a counting pass to figure out how many edges there will be
                            if (val_ptr) {
                                rowIdx_ptr[idx] = sub2ind (x1, y1, nrows);
                                colIdx_ptr[idx] = sub2ind (x2, y2, nrows);
                                
                                val_ptr[idx] = maximal_edge_entry (x1, y1, x2, y2, nrows, ncols, max_strength_ptr);
                            }
                            
                            ++idx;
                        }
                    }
                }
            }
        }
    }

    return idx;
}
                
/* The gateway function */
void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[]) 
{
    mwSize nrows, ncols, nedges;
    double radius, *max_strength_ptr;
    double *val_ptr, *rowIdx_ptr, *colIdx_ptr;
    
    
    // function signature:
    // [val, rowIdx, colIdx] = mxBuildWv4(max_strength, radius)
    
    /* check for proper number of arguments */
    if(nrhs!=2) {
        mexErrMsgIdAndTxt("mxBuildWv4:nrhs", "Two inputs required, max_strength and radius.");
    }
    
    if(nlhs!=3) {
        mexErrMsgIdAndTxt("mxBuildWv4:nlhs", "Three outputs required, val, rowIdx and colIdx.");
    }
    
    max_strength_ptr = checkArgDoubleArray(prhs, 0, "First argument must be an nrows x ncols array", 0, 0);
    
    nrows = mxGetM (prhs[0]);
    ncols = mxGetN (prhs[0]);
    
    radius = *checkArgDoubleArray(prhs, 1, "Second argument should be a scalar", 1, 1);
    
    if (radius <= 1) {
        mexErrMsgIdAndTxt("mxBuildWv4:radius", "Radius must be greater than 1");
    }
    
    // First pass to count the number of valid adjacencies
    nedges = InterveningContourStrength (nrows, ncols, max_strength_ptr, radius, NULL, NULL, NULL);
    
    // Allocate the output
    
    plhs[0] = mxCreateDoubleMatrix(nedges, 1, mxREAL);
    plhs[1] = mxCreateDoubleMatrix(nedges, 1, mxREAL);
    plhs[2] = mxCreateDoubleMatrix(nedges, 1, mxREAL);

    val_ptr    = mxGetPr(plhs[0]);
    rowIdx_ptr = mxGetPr(plhs[1]);
    colIdx_ptr = mxGetPr(plhs[2]);
    
    // Second pass to do the job
    nedges = InterveningContourStrength (nrows, ncols, max_strength_ptr, radius, val_ptr, rowIdx_ptr, colIdx_ptr);
}